import sys
import argparse
import asyncio

from .__init__ import __version__
from .cache import delete_from_cache, list_cache
from .webster import search_webster, get_webster_wod, get_webster_wod_list
from .camb import search_cambridge
from .utils import get_cache_selection, get_cache_selection_by_fzf


def parse_args(session):
    parser = argparse.ArgumentParser(
        description="Terminal Version of Cambridge Dictionary by default. Also supports Merriam-Webster Dictionary."
    )

    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="print the current version of the program",
    )

    parser.set_defaults(session=session)

    parser.add_argument(
        "--debug",
        action="store_true",
        help="turn on debug mode",
    )

    # Add sub-command capability that can identify different sub-command name
    sub_parsers = parser.add_subparsers(dest="subparser_name")

    # Add sub-command l
    parser_lw = sub_parsers.add_parser(
        "l",
        help="list alphabetically ordered words/phrases you've found before",
    )

    # Make sub-command l run default funtion of "list_words"
    parser_lw.set_defaults(func=list_words)

    # Add an optional argument for l command
    parser_lw.add_argument(
        "-d",
        "--delete",
        nargs="+",
        help="delete a word/phrase or multiple words/phrases(separated by ',') from cache",
    )

    # Add an optional argument for l command
    parser_lw.add_argument(
        "-t",
        "--time",
        action="store_true",
        help="list words/phrases you've found before in reverse chronological order",
    )

    # Add an optional argument for l command
    parser_lw.add_argument(
        "-r",
        "--random",
        action="store_true",
        help="randomly list 20 words/phrases you've found before",
    )

    # Add sub-command s
    parser_sw = sub_parsers.add_parser("s", help="look up words/phrases; hidden for convenience, no need to type")

    # Make sub-command s run default function of "search_words"
    parser_sw.set_defaults(func=search_word)

    # Add positional arguments with n args for s command
    parser_sw.add_argument(
        "-s",
        "--search",
        nargs="+",
        help="look up words/phrases (separated by ',') in Cambridge Dictionary",
    )

    # Add an optional argument for s command
    parser_sw.add_argument(
        "-c",
        "--chinese",
        nargs="+",
        help="look up words/phrases (separated by ',') in Cambridge Dictionary with Chinese translation",
    )

    # Add an optional argument for s command
    parser_sw.add_argument(
        "-w",
        "--webster",
        nargs="+",
        help="look up words/phrases (separated by ',') in Merriam-Webster Dictionary",
    )

    # Add an optional argument for s command
    parser_sw.add_argument(
        "-f",
        "--fresh",
        action="store_true",
        help="look up words/phrases afresh without using cache",
    )


    # Add an optional argument for s command
    parser_sw.add_argument(
        "-n",
        "--nosuggestions",
        action="store_true",
        help="look up words/phrases without showing spelling suggestions if not found",
    )

    # Add sub-command wod
    parser_wod = sub_parsers.add_parser(
        "wod",
        help="list today's Word of the Day from Merriam-Webster Dictionary",
    )

    # Make sub-command wod run default funtion of "wod"
    parser_wod.set_defaults(func=wod)

    # Add an optional argument for wod command
    parser_wod.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="list all words of the day",
    )

    if len(sys.argv) == 1:
        print_help(parser, parser_lw, parser_sw, parser_wod)
        sys.exit()

    elif sys.argv[1] == "-h" or sys.argv[1] == "--help":
        print_help(parser, parser_lw, parser_sw, parser_wod)
        sys.exit()

    elif sys.argv[1] == "-v" or sys.argv[1] == "--version":
        print("cambridge " + __version__)
        sys.exit()

    elif "l" in sys.argv[1 : ] or "wod" in sys.argv[1 : ]:
        args = parser.parse_args()
        return args

    else:
        argv_list = sys.argv[1 : ]
        w_index = None
        c_index = None
        s_index = None
        for index, word in enumerate(argv_list):
            if word == "--search" or word == "-s":
                s_index = index
            elif word == "--webster" or word == "-w":
                w_index = index
            elif word == "--chinese" or word == "-c":
                c_index = index
            elif word[0] != "-" and s_index is None and w_index is None and c_index is None:
                argv_list.insert(index, "-s")
                break
            else:
                continue

        to_parse = []
        for index, word in enumerate(argv_list):
            if word == "--debug":
                parser_sw.set_defaults(debug=True)
            elif word == "--fresh" or word == "-f":
                parser_sw.set_defaults(fresh=True)
            elif word == "--nosuggestions" or word == "-n":
                parser_sw.set_defaults(nosuggestions=True)
            elif (word == "--search" or word == "-s") and (index == len(argv_list) - 1 or argv_list[index + 1][0] == "-"):
                continue
            elif (word == "--webster" or word == "-w") and (index == len(argv_list) - 1 or argv_list[index + 1][0] == "-"):
                continue
            elif (word == "--chinese" or word == "-c") and (index == len(argv_list) - 1 or argv_list[index + 1][0] == "-"):
                continue
            else:
                to_parse.append(word)

        parser_sw.set_defaults(session=session)
        args= parser_sw.parse_args(to_parse)
        return args


def print_help(parser, parser_lw, parser_sw, parser_wod):
    parser.print_help()
    print("\n\n\033[1mCOMMAND l\033[0m")
    parser_lw.print_help()

    print("\n\n\033[1mCOMMAND s (hidden)\033[0m")
    parser_sw.print_help()

    print("\n\n\033[1mCOMMAND wod\033[0m")
    parser_wod.print_help()

    sys.exit()


async def list_words(args):
    if args.delete:
        to_delete = args.delete
        words = " ".join(to_delete)
        tasks = []
        for w in words.split(","):
            i = w.strip()
            if i:
                tasks.append(delete_from_cache(i))
        await asyncio.gather(*tasks)
        return

    if args.random:
        method = "random"
    elif args.time:
        method = "by_time"
    else:
        method = "by_alpha"

    has_fzf, data = list_cache(method)
    select_word = get_cache_selection_by_fzf(data) if has_fzf else get_cache_selection(data, method)

    if len(select_word) > 1 and not select_word.isnumeric():
        await search_webster(args.session, select_word)
    else:
        sys.exit()


async def search_word(args):
    if not args.search and not args.webster and not args.chinese:
        print("You didn't input any word or phrase.")
        sys.exit(3)

    tasks = []
    if args.search:
        cambridge_words = " ".join(args.search)
        for w in cambridge_words.split(","):
            i = w.strip(".").strip()
            if i:
                tasks.append(search_cambridge(args.session, i, args.fresh, False, args.nosuggestions, None))
    if args.webster:
        webster_words = " ".join(args.webster)
        for w in webster_words.split(","):
            i = w.strip(".").strip()
            if i:
                tasks.append(search_webster(args.session, i, args.fresh, args.nosuggestions, None))
    if args.chinese:
        words = " ".join(args.chinese)
        for w in words.split(","):
            i = w.strip(".").strip()
            if i:
                tasks.append(search_cambridge(args.session, i, args.fresh, True, args.nosuggestions, None))

    await asyncio.gather(*tasks)


async def wod(args):
    if args.list:
        await get_webster_wod_list(args.session)
    else:
        await get_webster_wod(args.session)



import sys
import sqlite3
from pathlib import Path

from .log import logger
from .utils import OP, has_tool, get_dict_name_by_url

dir = Path.home() / ".cache" / "cambridge"
dir.mkdir(parents=True, exist_ok=True)
DB = str(dir / "cambridge.db")

con = sqlite3.connect(DB, detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)


# NOTE Python sqslite3 syntax suger is bitter.

# The following functions suffixed with _table are simply DB operations and not complicated by higher level business logics
# The following functions suffixed with _cache are wrappers of functions suffixed with _table


def create_table():
    con.execute(
        """CREATE TABLE IF NOT EXISTS words (
        "input_word" TEXT NOT NULL,
        "response_word" TEXT NOT NULL,
        "created_at" TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
        "response_url" TEXT UNIQUE NOT NULL,
        "response_text" TEXT NOT NULL,
        UNIQUE(response_url))"""
    )


def insert_entry_into_table(input_word, response_word, url, text):
    import datetime
    current_datetime = datetime.datetime.now()

    try:
        # 1. About sqlite3 context manager (from sqlite3 — DB-API 2.0 interface for SQLite databases):
        # "A Connection object can be used as a context manager that automatically commits or rolls back open transactions
        # when leaving the body of the context manager. If the body of the with statement finishes without exceptions, the transaction is committed."
        # "Connection object used as context manager **ONLY** commits or rollbacks transactions, so the connection object should be closed manually."
        # NOTE Assuming a context manager is designed in the first place for automatic close operation is wrong in the case of sqlite connection object.

        # 2. About how to use sqlite3 "RETURN" primitive in Python:
        # According to 1, Python autocommits if you use context manager and when the code leaves the with block, and you later fetchone(), RETURN won't work and you will get:
        # OperationalError: cannot commit transaction - SQL statements in progress
        # According to sqlite3 docs, only after SQLITE_DONE, the RETURN starts to work.
        # Then the code is dead with the context manager: COMMIT (when leaving with block) depends on RETURN, RETURN depends on SQLITE_DONE, SQLITE_DONE happens after COMMIT
        # NOTE Don't use a connection as context manager when RETURN is needed.

        # 4. After digging deep into cpython/modules/_sqlite/cursor.c, Python deals with sqlite3 step SQLITE_DONE within fetchone(),
        # fetchone() first, then manual COMMIT, and RETURN will work.

        # 5. ON CONFLICT only takes effect when there is PRIMARY KEY or UNIQUE constraint present in the table schema.
        # You can't enforce it with other column names, or it will throw:
        # OperationalError: ON CONFLICT clause does not match any PRIMARY KEY or UNIQUE constraint
        # Example code demonstrating UPSERT i.e. ON CONFLICT (column_name) DO UPDATE SET (The code in use chooses DO NOTHING conflict resolution instead):
        """
        con.execute(
            "INSERT INTO words (input_word, response_word, created_at, response_url, response_text)
            VALUES (?, ?, ?, ?, ?)
            ON CONFLICT (input_word) DO UPDATE SET
            response_word=excluded.response_word,
            created_at=excluded.created_at,
            response_url=excluded.response_url,
            response_text=excluded.response_text
            ",
            (input_word, response_word, current_datetime, url, text)
        )
        """
        res_word = con.execute(
            """INSERT INTO words (input_word, response_word, created_at, response_url, response_text)
            VALUES (?, ?, ?, ?, ?)
            ON CONFLICT DO NOTHING
            RETURNING response_word
            """,
            (input_word, response_word, current_datetime, url, text)
        ).fetchone()

    except sqlite3.OperationalError as error:
        if "no such table" in str(error):
            create_table()
        else:
            raise
    except sqlite3.Error:
        raise
    else:
        con.commit()
        return res_word


def check_word_in_table(word, request_url):
    try:
        cur = con.execute(
            "SELECT response_url FROM words WHERE response_url = ? OR response_word = ? OR input_word = ?",
            (request_url, word, word),
        )
    except sqlite3.OperationalError as error:
        if "no such table" in str(error):
            create_table()
        else:
            raise
    except sqlite3.Error:
        raise
    else:
        return cur.fetchone()


def get_entry_from_table(response_url):
    try:

        # From https://docs.python.org/2/library/
        # "For the qmark style, parameters must be a sequence whose length must match the number of placeholders, or a ProgrammingError is raised."
        # sqlite3.ProgrammingError: Incorrect number of bindings supplied. The current statement uses 1, and there are 57 supplied.
        # "Note that it is actually the comma which makes a tuple, not the parentheses.
        # The parentheses are optional, except in the empty tuple case, or when they are needed to avoid syntactic ambiguity.
        # For example, f(a, b, c) is a function call with three arguments, while f((a, b, c)) is a function call with a 3-tuple as the sole argument."
        # NOTE (response_url) without comma won't be treated as sequence. (response_url,) should be used here.

        cur = con.execute(
            "SELECT response_word, response_text FROM words WHERE response_url = ?",
            (response_url,),
        )
    except sqlite3.Error:
        raise
    else:
        return cur.fetchone()


def get_entries_from_table(is_random=False):
    try:
        if is_random:
            cur = con.execute("SELECT response_word, response_url FROM words ORDER BY RANDOM() LIMIT 20")
        else:
            cur = con.execute("SELECT response_word, response_url, created_at FROM words")
    except sqlite3.OperationalError as error:
        if "no such table" in str(error):
            create_table()
        else:
            raise
    except sqlite3.Error:
        raise
    else:
        return cur.fetchall()


def delete_entry_from_table(word):
    try:
        res_url = con.execute("DELETE FROM words WHERE response_word = ? OR input_word = ? RETURNING response_url", (word, word)).fetchone()
    except sqlite3.OperationalError as error:
        if "no such table" in str(error):
            create_table()
        else:
            raise
    except sqlite3.Error:
        raise
    else:
        con.commit()
        return res_url


async def delete_from_cache(word):
    try:
        result = delete_entry_from_table(word)
    except sqlite3.Error as error:
        logger.error(f'{OP.CANCELLED.name} deleting "{word}" from cache: [{error.__class__.__name__}] {error}\n')
        sys.exit(4)
    else:
        if result is None:
            print(f'{OP.NOT_FOUND.name} "{word}" in cache')
        else:
            dict_name = get_dict_name_by_url(result[0])
            print(f'{OP.DELETED.name} "{word}" from {dict_name} in cache successfully')


def list_cache(method):
    is_random = True if method == "random" else False

    try:
        data = get_entries_from_table(is_random)
    except sqlite3.Error as error:
        logger.error(f'{OP.CANCELLED.name} listing cache: [{error.__class__.__name__}] {error}\n')
        sys.exit(4)
    else:
        if data is None:
            print("You may haven't searched any word yet")
            sys.exit(3)

        has_fzf = has_tool("fzf")
        if method == "by_alpha":
            data.sort()
        elif method == "by_time":
            reverse = True if has_fzf else False
            data.sort(reverse=reverse, key=lambda tup: tup[2])

        return (has_fzf, data)


async def check_cache(input_word, req_url):
    logger.debug(f'{OP.SEARCHING.name} "{input_word}" in cache')

    try:
        data = check_word_in_table(input_word, req_url)
    except sqlite3.Error as error:
        logger.error(f'{OP.CANCELLED.name} searching "{input_word}" in cache: [{error.__class__.__name__}] {error}\n')
        sys.exit(4)
    else:
        # considering user might input plural nouns, or verbs with tenses
        if data is None:
            if "s" != input_word[-1]:
                return None
            else:
                data = check_word_in_table(input_word[:-1], req_url)
                if data is None:
                    if "es" != input_word[-2:]:
                        return None
                    else:
                        data = check_word_in_table(input_word[:-2], req_url)
                        if data is None:
                            return None
        return data[0]


async def get_cache(response_url):
    result = get_entry_from_table(response_url)
    if result is None:
        logger.error(f'The function argument "{response_url}" must actually come from the cache.')
        sys.exit(4)
    return result


async def save_to_cache(input_word, response_word, response_url, response_text):
    try:
        result = insert_entry_into_table(input_word, response_word, response_url, response_text)

        # a duplicate check way other than ON CONFLICT
        # sqllite3.IntegrityError: NOT NULL constraint failed
        # sqllite3.IntegrityError: UNIQUE constraint failed
        # Only with `as error`, can you get str(error) as error message; str(sqlite3.IntegrityError) gives you "<class 'sqlite3.IntegrityError'>".
        """
        except sqlite3.IntegrityError as error:
            if "UNIQUE constraint" in str(error): logger.debug(f'{OP.CANCELLED.name} caching "{input_word}": {error}\n')
        """

    except sqlite3.Error as error:
        logger.error(f'{OP.CANCELLED.name} caching "{input_word}": [{error.__class__.__name__}] {error}\n')
        sys.exit(4)

    else:
        if result is None:
            logger.debug(f'{OP.CANCELLED.name} caching, already cached before') # hit ON CONFLICT
        else:
            logger.debug(f'{OP.CACHED.name} "{result[0]}"')



import sys
import subprocess
import asyncio
from urllib import parse
from enum import Enum
from fake_user_agent import aio_user_agent # type: ignore

from .log import logger
from .console import c_print

from typing import Optional, Literal
Initiator = Literal["wod_calendar", "spell_check", "cache_list", "redirect_list"]


class OP(Enum):
    FETCHING        = 1,
    PARSING         = 2,
    RETRY_FETCHING  = 3,
    RETRY_PARSING   = 4,
    PRINTING        = 5,
    FOUND           = 6,
    NOT_FOUND       = 7,
    CACHED          = 8,
    CANCELLED       = 9,
    CANCELLING      = 10,
    DELETED         = 11,
    UPDATED         = 12,
    SWITCHED        = 13,
    SEARCHING       = 14,
    SELECTED        = 15


class DICT(Enum):
    CAMBRIDGE       = 1,
    MERRIAM_WEBSTER = 2


def get_dict_name_by_url(url):
    return DICT.CAMBRIDGE.name if DICT.CAMBRIDGE.name.lower() in url else DICT.MERRIAM_WEBSTER.name


def quit_on_error(path, error, op):
    print(f'{op} <{path}> failed: [{error.__class__.__name__}] {error}')
    sys.exit(2)


def quit_on_no_result(dict_name, is_spellcheck=False):
    w = "result" if not is_spellcheck else "suggestions"
    print(f"No {w} found in {dict_name}")
    sys.exit(1)


def cancel_on_error(path, error, attempt, op):
    attempt += 1
    logger.debug(f"{op} {path} {attempt} times")

    if attempt == 3:
        print(f'Maximum {op} reached.')
        cancel_on_error_without_retry(path, error, op)

    return attempt


def cancel_on_error_without_retry(path, error, op):
    print(f'{op} on {path} failed: [{error.__class__.__name__}]')

    logger.debug(f"{OP.CANCELLING.name} task...")
    task = asyncio.current_task()
    task.cancel() # type: ignore


async def fetch(session, url):
    attempt = 0
    ua = await aio_user_agent()
    logger.debug(f"Got User-Agent: {ua}")
    logger.debug(f"{OP.FETCHING.name} {url}")
    while True:
        try:
            resp = await session.get(url, headers={"User-Agent": ua}, timeout=5, ssl=False)
        except asyncio.TimeoutError as error:
            attempt = cancel_on_error(url, error, attempt, OP.FETCHING.name)
            continue
        except Exception as error:
            cancel_on_error_without_retry(url, error, OP.FETCHING.name) # NO break!
        else:
            return resp


def replace_all(string):
    return (
        string.replace("\n            (", "(")
        .replace("(\n    \t                ", "(")
        .replace("\n", " ")
        .replace("      \t                ", " ")
        .replace("\\'", "'")
        .replace("  ", " ")
        .replace("\xa0 ", "")
        .replace("[ ", "")
        .replace(" ]", "")
        .replace("[", "")
        .replace("]", "")
        .replace("A1", "")
        .replace("A2", "")
        .replace("B1", "")
        .replace("B2", "")
        .replace("C1", "")
        .replace("C2", "")
        .strip()
    )


def parse_response_url(url):
    return url.split("?")[0]


def get_request_url(url, input_word, dict):
    if dict == DICT.CAMBRIDGE.name:
        if " " in input_word:
            query_word = input_word.replace(" ", "+").replace("/", "+")
            request_url = url + query_word
        else:
            request_url = url + input_word
        return request_url

    return url + parse.quote(input_word)


def decode_url(url):
    return parse.unquote(url)


def print_word_per_line(index, word, extra=""):
    import os
    cols = os.get_terminal_size().columns
    word_len = len(word)

    if index % 2 == 0:
        print(f"\033[37;1;100m{index+1:6d}|{word}", end="")
        print(f"\033[37;1;100m{extra:>{cols-word_len-7}}\033[0m")
    else:
        c_print(f"#[bold #4A7D95]{index+1:6d}|{word}", end="")
        c_print(f"#[bold #4A7D95]{extra:>{cols-word_len-7}}")


def list_items(data, initiator: Optional[Initiator] = None):
    for index, entry in enumerate(data):
        if initiator == "cache_list":
            word = entry[0]
            dict_name = get_dict_name_by_url(entry[1])
            print_word_per_line(index, word, dict_name)
        elif initiator == "wod_calendar":
            date_string = data[entry].split("/")[-1]
            date = ""
            for i, c in enumerate(date_string):
                if c.isnumeric():
                    date = date_string[i:]
                    break
            print_word_per_line(index, entry, date)
        else:
            print_word_per_line(index, entry, "")


def has_tool(name):
    from shutil import which
    return which(name) is not None


def get_suggestion_notice(dict_name, has_fzf):
    flip_dict = DICT.MERRIAM_WEBSTER.name if (dict_name == DICT.CAMBRIDGE.name) else DICT.CAMBRIDGE.name
    if has_fzf:
        return f"[ENTER] to print the selected suggestion's meaning; Any [NUMBER] to switch to {flip_dict}, or a new word to search; [ESC] to quit out."
    else:
        return f"[ENTER] to switch to {flip_dict}; [NUMBER] to print the selected suggestion's meaning, or a new word to search; Any Other [KEY] to quit out: "


def get_suggestion(suggestions, dict_name):
    c_print(dict_name + (int(len(dict_name)/2))*" ", justify="center")
    list_items(suggestions, "spell_check")
    print(get_suggestion_notice(dict_name, has_fzf=False), end="")
    key = input("")

    if (key.isnumeric() and (1 <= int(key) <= len(suggestions))):
        return suggestions[int(key) - 1]
    elif len(key) > 1:
        return key
    elif key == "":
        return ""
    else:
        sys.exit()


def get_suggestion_by_fzf(suggestions, dict_name):
    notice = get_suggestion_notice(dict_name, has_fzf=True)
    choices = "\n".join(suggestions).strip("\n")
    p1 = subprocess.Popen(["echo", choices], stdout=subprocess.PIPE, text=True)
    p2 = subprocess.Popen(["fzf", "--layout=reverse", "--bind", "enter:accept-or-print-query", "--header", notice], stdin=p1.stdout, stdout=subprocess.PIPE, text=True)
    select_word = p2.communicate()[0].strip("\n")

    if len(select_word) == 1 and select_word.isnumeric():
        return ""
    elif p2.returncode == 0 and select_word != "":
        return select_word
    else:
        sys.exit()


def get_wod_selection(data):
    title = DICT.MERRIAM_WEBSTER.name + " Calendar of Word of the Day"
    c_print(title + (int(len(title)/2))*" ", justify="center")
    list_items(data, "wod_calendar")
    notice =  "Type one word of day from above to print the meaning; Input a new word; [ANY OTHER KEY] to quit out: "
    print(notice, end="")
    return input("")


def get_wod_selection_by_fzf(data):
    notice = "Select to print the word-of-the-day meaning; [ESC] to quit out."
    choices = "\n".join(data.keys()).strip("\n")

    p1 = subprocess.Popen(["echo", choices], stdout=subprocess.PIPE, text=True)
    p2 = subprocess.Popen(["fzf", "--layout=reverse", "--bind", "enter:accept-or-print-query", "--header", notice], stdin=p1.stdout, stdout=subprocess.PIPE, text=True)
    select_word = p2.communicate()[0].strip("\n")

    if p2.returncode == 0 and select_word != "":
        return select_word
    else:
        sys.exit()


def get_cache_selection(data, method):
    title = "List of Cache" + f" ({method})"
    c_print(title + (int(len(title)/2))*" ", justify="center")
    list_items(data, "cache_list")
    notice =  "Type one word from above to print the meaning; Input a new word; [ANY OTHER KEY] to quit out: "
    print(notice, end="")
    return input("")


def get_cache_selection_by_fzf(data):
    notice = "Select to print the word's meaning; [ESC] to quit out."
    choices = ""
    for i in data:
        choices = choices + i[0] + "\n"
    choices = choices.strip()

    p1 = subprocess.Popen(["echo", choices], stdout=subprocess.PIPE, text=True)
    p2 = subprocess.Popen(["fzf", "--layout=reverse", "--bind", "enter:accept-or-print-query", "--header", notice], stdin=p1.stdout, stdout=subprocess.PIPE, text=True)
    select_word = p2.communicate()[0].strip("\n")

    if p2.returncode == 0 and select_word != "":
        return select_word
    else:
        sys.exit()



import os
import re

from .color import COLOR_EFFECT

from typing import (
    Optional,
    Literal
)

JustifyMethod = Literal["left", "center", "right"]

Symbol = {
    "L_BRACKET" : "[",
    "R_BRACKET" : "]",
    "SLASH"     : "/",
    "HASH"      : "#"
}


def hex_to_rgb(hex):
    h = hex[1:]
    return tuple(int(h[i : i + 2], 16) for i in (0, 2, 4))


def get_color_escape(r, g, b, background=False):
    return '\033[{};2;{};{};{}m'.format(48 if background else 38, r, g, b)


def get_color_effect(code):
    return '\033[' + COLOR_EFFECT[code] + 'm'


def parse_in_bracket(text):
    new_text = ""

    for i, t in enumerate(text):
        if t == Symbol["SLASH"]:
            new_text = get_color_effect("RESET")
            return new_text

        if t == Symbol["HASH"]:
            new_text += get_color_escape(*hex_to_rgb(text[i : i + 7]), background=False)

    for ce in COLOR_EFFECT.keys():
        if ce.lower() in text:
            new_text += get_color_effect(ce)

    return new_text


def parse(string):
    texts = re.split(r'[\[^[\]]', string)

    if len(texts) == 1:
        return texts[0]

    length = len(string)
    after_parse = ""

    i = 0
    while i < length:
        s = string[i]

        if s not in Symbol.values():
            after_parse += s
            i = i + 1
        elif s == Symbol["SLASH"] and string[i - 1] != Symbol["L_BRACKET"]:
            after_parse += s
            i = i + 1
        elif s == Symbol["HASH"] and string[i + 1] != Symbol["L_BRACKET"]:
            after_parse += s
            i = i + 1
        elif s == Symbol["HASH"] and string[i + 1] == Symbol["L_BRACKET"]:
            i = i + 1
        elif s == Symbol["L_BRACKET"] and string[i - 1] == Symbol["HASH"]:
                k = i
                for j, ss in enumerate(string[i + 1 : ]):
                    k += 1
                    if ss == Symbol["R_BRACKET"]:
                        text_in_bracket = string[i + 1 : i + 1 + j]
                        for word in text_in_bracket.split():
                            after_parse += parse_in_bracket(string[i + 1 : i + 1 + j])
                            i = k
                            break
                        break
                i = i + 1

        elif s == Symbol["L_BRACKET"] and string[i - 1] != Symbol["HASH"]:
            after_parse += s
            i = i + 1
        else:
            after_parse += s
            i = i + 1

    return after_parse + get_color_effect("RESET")


def c_print(*objects, sep=' ', end='\n', file=None, flush=False, justify: Optional[JustifyMethod] = None):
    if not objects:
        objects = ("\n",)

    if len(objects)  == 1:
        text = parse(objects[0])
    else:
        new_data = ""
        for i in objects:
            new_data = new_data + i
        text = parse(new_data)

    if justify is not None and isinstance(objects[0], str):
        cols = os.get_terminal_size().columns

        # https://docs.python.org/3/library/string.html#grammar-token-format-spec-align
        # FIXME to strip out color effect characters when justifying
        if justify == "right":
            print(f"{text:>{cols}}")
        elif justify == "left":
            print(f"{text:<{cols}}")
        elif justify == "center":
            print(f"{text:^{cols}}")
        else:
            justify = None
    else:
        print(text, end=end)



COLOR_EFFECT = {
    "BLACK"         : "30",
    "RED"           : "31",
    "GREEN"         : "32",
    "YELLOW"        : "33",
    "BLUE"          : "34",
    "MAGENTA"       : "35",
    "CYAN"          : "36",
    "WHITE"         : "37",
    "RESET"         : "0",
    "BOLD"          : "1",
    "ITALIC"        : "3"
}

# Webster
accessory = "#4A7D95"
badge = "#E06C75"
or_badge = "#757575"

# 1. Webster dictionary-entry
# 1.1 entry-header-content
eh_h1_word = "#2E8B57"
eh_word_type = "#FD7E14"
eh_word_syllables = "#757575"
# eh_pron = "#4A7D95" # UNUSED

# 1.2 verb tenses, noun plurals, adjective comparative forms
# wt = "#525A5B" # UNUSED

# 1.3 other word forms
wf = "#4A7D95"
wf_type = "#4A7D95"

# 1.4 transitive / intransitive
# tran = "yellow" # UNUSED

# 1.5 one meaning of one entry
meaning_num = "#757575"
meaning_letter = "#757575"
meaning_link = "#4A7D95"
meaning_content = "#4A7D95"
meaning_badge = badge
meaning_arrow = "#BC8F8F"
meaning_keyword = "#757575"
meaning_sentence = "#757575"

### 1.6 dxnls
dxnls_title = "#2E8B57"
dxnls_content = "#757575"

# 2. Webster phrases
ph_title = "#2E8B57"
ph_item = "#4A7D95"

# 3. Webster synonyms
syn_title = ph_title
syn_item = "#4A7D95"
syn_em = "#2E8B57"
# syn_label = "#0F3850"
syn_label = "#757575"

# 4. Webster examples on the web
eg_title = ph_title
eg_sentence = "#757575"
eg_word = "#2E8B57"

# 5. Webster related phrases
rph_title = ph_title
rph_em = "#2E8B57"
rph_item = "#4A7D95"

# 6. Webster nearby entries
nearby_title = ph_title
nearby_em = "#2E8B57"
nearby_word = "#757575"
nearby_item = "#4A7D95"

# 7. Webster dict name
dict_name = "#757575"

# 8. Webster Word of the Day
wod_title = ph_title
wod_type = eh_word_type
wod_syllables = eh_word_syllables
wod_subtitle = wod_title
wod_sent = meaning_sentence



import sys
import re
from bs4 import BeautifulSoup # type: ignore
import asyncio

from .console import c_print
from .log import logger
from .utils import fetch, get_request_url, parse_response_url, replace_all, OP, DICT, has_tool, get_suggestion, get_suggestion_by_fzf, cancel_on_error, quit_on_no_result, cancel_on_error_without_retry
from .cache import check_cache, save_to_cache, get_cache
from . import webster


CAMBRIDGE_URL = "https://dictionary.cambridge.org"
CAMBRIDGE_EN_SEARCH_URL = CAMBRIDGE_URL + "/search/direct/?datasetsearch=english&q="
CAMBRIDGE_CN_SEARCH_URL = CAMBRIDGE_URL + "/search/direct/?datasetsearch=english-chinese-simplified&q="

CAMBRIDGE_SPELLCHECK_URL = CAMBRIDGE_URL + "/spellcheck/english/?q="
CAMBRIDGE_SPELLCHECK_URL_CN = CAMBRIDGE_URL + "/spellcheck/english-chinese-simplified/?q="


async def search_cambridge(session, input_word, is_fresh=False, is_ch=False, no_suggestions=False, req_url=None):
    if req_url is None:
        url = CAMBRIDGE_CN_SEARCH_URL if is_ch else CAMBRIDGE_EN_SEARCH_URL
        req_url = get_request_url(url, input_word, DICT.CAMBRIDGE.name)

    if is_fresh:
        await fresh_run(session, input_word, is_ch, no_suggestions, req_url)
    else:
        res_url = await check_cache(input_word, req_url)
        if res_url is None:
            logger.debug(f'{OP.NOT_FOUND.name} "{input_word}" in cache')
            await fresh_run(session, input_word, is_ch, no_suggestions, req_url)
        elif DICT.CAMBRIDGE.name.lower() not in res_url:
            await webster.cache_run(res_url)
        else:
            await cache_run(res_url)


async def cache_run(res_url_from_cache):
    res_word, res_text = await get_cache(res_url_from_cache)
    logger.debug(f'{OP.FOUND.name} "{res_word}" from {DICT.CAMBRIDGE.name} in cache')
    logger.debug(f"{OP.PARSING.name} {res_url_from_cache}")

    soup = BeautifulSoup(res_text, "lxml")
    first_dict = soup.find("div", "pr dictionary") or soup.find("div", "pr di superentry")

    logger.debug(f"{OP.PRINTING.name} the parsed result of {res_url_from_cache}")
    blocks = first_dict.find_all("div", ["pr entry-body__el", "entry-body__el clrd js-share-holder", "pr idiom-block"]) # type: ignore
    for block in blocks:
        parse_dict_head(block)
        parse_dict_body(block)
    c_print(f'\n#[#757575]{OP.FOUND.name} "{res_word}" from {DICT.CAMBRIDGE.name} in cache. You can add "-f -w" to fetch the {DICT.MERRIAM_WEBSTER.name} dictionary')


async def fresh_run(session, input_word, is_ch, no_suggestions, req_url):
        response = await fetch(session, req_url)
        res_url = str(response.real_url)

        if "spellcheck" in res_url:
            logger.debug(f'{OP.NOT_FOUND.name} "{input_word}" at {req_url}')

            if no_suggestions:
                sys.exit(-1)

            spell_base_url = CAMBRIDGE_SPELLCHECK_URL_CN if is_ch else CAMBRIDGE_SPELLCHECK_URL
            spell_req_url = get_request_url(spell_base_url, input_word, DICT.CAMBRIDGE.name)

            spell_res = await fetch(session, spell_req_url)
            spell_res_text = None
            attempt = 0
            while True:
                try:
                    spell_res_text = await spell_res.text()
                except asyncio.TimeoutError as error:
                    attempt = cancel_on_error(spell_req_url, error, attempt, OP.FETCHING.name)
                    continue
                except Exception as error:
                    cancel_on_error_without_retry(spell_req_url, error, OP.FETCHING.name)
                    break
                else:
                    break

            if spell_res_text is not None:
                logger.debug(f"{OP.PARSING.name} out suggestions at {spell_res.url}")
                soup = BeautifulSoup(spell_res_text, "lxml")
                node = soup.find("div", "hfl-s lt2b lmt-10 lmb-25 lp-s_r-20")
                suggestions = []

                if not node:
                    quit_on_no_result(DICT.CAMBRIDGE.name, is_spellcheck=True)

                for ul in node.find_all("ul", "hul-u"): # type: ignore
                    if "We have these words with similar spellings or pronunciations:" in ul.find_previous_sibling().text:
                        for i in ul.find_all("li"):
                            sug = replace_all(i.text)
                            suggestions.append(sug)

                logger.debug(f"{OP.PRINTING.name} out suggestions at {spell_res.url}")
                select_word = get_suggestion_by_fzf(suggestions, DICT.CAMBRIDGE.name) if has_tool("fzf") else get_suggestion(suggestions, DICT.CAMBRIDGE.name)
                if select_word == "":
                    logger.debug(f'{OP.SWITCHED.name} to {DICT.MERRIAM_WEBSTER.name}')
                    await webster.search_webster(session, input_word, True, no_suggestions, None) # type: ignore
                else:
                    logger.debug(f'{OP.SELECTED.name} "{select_word}"')
                    await search_cambridge(session, select_word, False, False, no_suggestions, None)

        else:
            res_url = parse_response_url(res_url)
            res_text = None
            attempt = 0
            while True:
                try:
                    res_text = await response.text()
                except asyncio.TimeoutError as error:
                    attempt = cancel_on_error(req_url, error, attempt, OP.FETCHING.name)
                    continue
                # There is also a scenario that in the process of cancelling, while is still looping, leading to run coroutine with a closed session.
                # If session is closed, and you go on connecting, ClientConnectionError will be throwed.
                except Exception as error:
                    cancel_on_error_without_retry(req_url, error, OP.FETCHING.name)
                    break
                else:
                    break

            if res_text is not None:
                soup = BeautifulSoup(res_text, "lxml")
                first_dict = soup.find("div", "pr dictionary") or soup.find("div", "pr di superentry")
                if first_dict is None:
                    quit_on_no_result(DICT.CAMBRIDGE.name, is_spellcheck=False)

                logger.debug(f'{OP.FOUND.name} "{input_word}" in {DICT.CAMBRIDGE.name} at {req_url}')
                logger.debug(f"{OP.PARSING.name} {req_url}")
                blocks = first_dict.find_all("div", ["pr entry-body__el", "entry-body__el clrd js-share-holder", "pr idiom-block"]) # type: ignore
                if len(blocks) == 0:
                    quit_on_no_result(DICT.CAMBRIDGE.name, is_spellcheck=False)
                else:
                    logger.debug(f"{OP.PRINTING.name} the parsed result of {req_url}")
                    for block in blocks:
                        parse_dict_head(block)
                        parse_dict_body(block)

                    print()

                    hword = soup.find("b", "tb ttn").text # type: ignore
                    await save_to_cache(input_word, hword, res_url, str(first_dict))


def parse_dict_head(block):
    head = block.find("div", "pos-header dpos-h")
    word_block = block.find("div", "di-title") or head.find("div", "di-title")
    hword = word_block.text

    if head is None:
        c_print("\n#[bold blue]" + hword, end="")

        info = block.find_all("span", ["pos dpos", "lab dlab", "v dv lmr-0"])
        if len(info) != 0:
            temp = [i.text for i in info]
            type = temp[0]
            text = " ".join(temp[1:])
            print(f" {type} {text}")
        else:
            print()

    else:
        vars = head.find_all("span", "var dvar")
        spellvar = head.find("span", "spellvar dspellvar")
        irreg = head.find("span", "irreg-infls dinfls")
        domain = head.find("span", "domain ddomain")

        dlab = None
        lab = head.find("span", "lab dlab")
        if lab is not None:
            lab_parent = head.find("span", "lab dlab").find_parent()
            if lab_parent.has_attr('class') and lab_parent['class'][0] == "pos-header":
                dlab = lab

        w_type = ""
        next_sibling = None

        dpos = head.find(title=re.compile("A word that"))
        if dpos is not None:
            w_type += dpos.text
            dgram = dpos.find_next_sibling("span", "gram dgram")
            if dgram is not None:
                w_type += " " + dgram.text
            w_type = w_type.strip("\n").strip().replace(" or ", "/")

            parent = dpos.find_parent()
            if parent['class'][-1] == "lmr-5":
                next_sibling = parent.find_next_sibling()
            else:
                if dgram is not None:
                    next_sibling = dgram.find_next_sibling()
                else:
                    next_sibling = dpos.find_next_sibling()

        if next_sibling is None:
            end = ""
        elif next_sibling is not None and next_sibling.has_attr('class') and next_sibling['class'][0] == "lml--5":
            end = " "
        elif len(vars) > 0 or spellvar is not None or domain is not None or dlab is not None or irreg is not None:
            end = " "
        else:
            end = "\n"
        c_print(f"\n#[bold blue]{hword}#[/bold blue] #[bold yellow]{w_type}#[/bold yellow]", end=end)

        if domain is not None:
            print(domain.text.strip("\n").strip(), end=" ")

        if dlab is not None:
            next_sibling = dlab.find_next_sibling()
            end = " " if ((next_sibling is not None and next_sibling.has_attr('class')) or len(vars) > 0 or irreg is not None) else "\n"
            print(dlab.text.strip("\n").strip(), end=end)

        if len(vars) > 0:
            for var in vars:
                next_sibling = var.find_next_sibling()
                end = " " if ((next_sibling is not None and next_sibling.has_attr('class')) or irreg is not None) else "\n"
                print(var.text.replace("\n", "").replace("Your browser doesn't support HTML5 audio", " ").replace("/", "|").strip(), end=end)

        if spellvar is not None:
            next_sibling = spellvar.find_next_sibling()
            end = " " if next_sibling is not None and next_sibling.has_attr('class') else "\n"
            print(spellvar.text.strip("\n").strip(), end=end)

        if irreg is not None:
            infgroups = irreg.find_all("span", "inf-group dinfg")
            if not infgroups:
                print(irreg.text.strip("\n").strip())
            else:
                # intentionally leave out pronunciations of the plural form:
                # e.g. cortex: uk  /ˈkɔː.tɪ.siːz/ us  /ˈkɔːr.tɪ.siːz/
                # e.g. vortex: uk  /-tɪ.siːz/ us  /-tə-/
                for index, infgroup in enumerate(infgroups):
                    infdlab = infgroup.find("span", "lab dlab")
                    if infdlab is not None:
                        previous_sibling = infdlab.find_previous_sibling()
                        if previous_sibling is not None:
                            c_print(f"#[bold] {previous_sibling.text}#[/bold]", end=" or ")

                        print(infdlab.text, end="")

                        next_sibling = infdlab.find_next_sibling()
                        if next_sibling is not None:
                            c_print(f"#[bold] {next_sibling.text}#[/bold]", end="")

                    else:
                        b = infgroup.find("b")
                        if b is not None:
                            next = b.find_next_sibling()
                            end = " or " if next is not None else ""
                            c_print(f"#[bold]{b.text}#[/bold]", end=end)
                            if next is not None and next.has_attr('class') and next['class'][0] == "inf":
                                c_print(f"#[bold]{next.text}#[/bold]", end="")

                    if index != len(infgroups) - 1:
                        print(" | ", end="")

                print()


        prons = head.find_all("span", "pron dpron")
        if len(prons) != 0:
            for pron in prons:
                parent = pron.find_parent()
                if parent.find_parent()['class'][0] == "pos-header":
                    pron_text = pron.text.strip("\n").strip().replace("/", "|")
                    area = parent.find("span", "region dreg")
                    area_text = area.text if area is not None else ""
                    end= "" if parent.find_next_sibling() is None else " "
                    c_print(f"#[bold]{area_text} #[/bold]" + pron_text, end=end)

            print()


def parse_def_title(block):
    d_title = replace_all(block.find("h3", "dsense_h").text)
    c_print("#[red]" + "\n" + d_title.upper())


def parse_ptitle(block):
    p_title = block.find("span", "phrase-title dphrase-title").text
    p_info = block.find("span", "phrase-info dphrase-info")

    if p_info is not None:
        phrase_info = replace_all(p_info.text)
        print(f"\033[34;1m  {p_title}\033[0m \033[33;1m{phrase_info}\033[0m")
    else:
        print(f"\033[34;1m  {p_title}\033[0m")


def parse_meaning(def_block, is_pmeaning=False):
    if is_pmeaning:
        print("  ", end="")

    meaning_b = def_block.find("div", "def ddef_d db")
    usage_b = meaning_b.find("span", "lab dlab")
    if usage_b is not None:
        usage = replace_all(usage_b.text)
        meaning_words = replace_all(meaning_b.text).split(usage)[-1]
        if meaning_words[-1] == ":":
            meaning_words = meaning_words[ : -1]
        print("\033[34;1m: \033[0m" + "[" + usage + "] " + "\033[34m" + meaning_words.strip() + "\033[0m", end="")
    else:
        meaning_words = replace_all(meaning_b.text)
        if meaning_words[-1] == ":":
            meaning_words = meaning_words[ : -1]

        print("\033[34;1m: \033[0m" + "\033[34m" + meaning_words.strip() + "\033[0m", end="")

    # e.g. def info tags like 'B1 [ C or U ]'
    def_info = replace_all(def_block.find("span", "def-info ddef-info").text).replace(" or ", "/")
    if def_info:
        if not def_info.startswith("("):
            print(" [" + def_info + "]", end="")
        else:
            print(" " + def_info, end="")

    # print the meaning's specific language translation if any
    meaning_lan = def_block.find("span", "trans dtrans dtrans-se break-cj")
    if meaning_lan is not None:
        meaning_lan_words = meaning_lan.text.replace(";", "；").replace(",", "，")
        if not meaning_lan_words.startswith("（"):
            print(" ", end="")
        print("\033[34m" + meaning_lan_words + "\033[0m")
    else:
        print()


def print_example_tag(tag_block):
    if tag_block is None:
        return

    tag = tag_block.text
    tag = replace_all(tag)
    print("[" + tag + "]", end= " ")


def parse_example(def_block, is_pexample=False):
    es = def_block.find_all("div", "examp dexamp")
    if len(es) == 0:
        return

    for e in es:
        result = e.find("span", "eg deg")
        if result is not None:
            example = replace_all(result.text)
        else:
            continue

        if is_pexample:
            print("  ", end="")

        example_lan = e.find("span", "trans dtrans dtrans-se hdb break-cj")
        if example_lan is not None:
            example_lan_sent = " " + example_lan.text.replace("。", "")
        else:
            example_lan_sent = ""

        c_print("#[blue]" + "|" + "#[/blue]", end="")

        dlab = e.find("span", "lab dlab")
        print_example_tag(dlab)

        dgram = e.find("span", "gram dgram")
        print_example_tag(dgram)

        dlu = e.find("span", "lu dlu")
        print_example_tag(dlu)

        c_print(f"#[#757575]{example}{example_lan_sent}#[/#757575]")


def print_tag(node):
    for i in node.find_all("span"):
        has_class = i.has_attr("class")
        parent = i.find_parent()
        parent_has_class = parent.has_attr("class")

        if has_class and i.attrs["class"] == ['x-h', 'dx-h']:
            c_print("#[#757575]" + "  • " + i.text, end="")
        elif has_class and i.attrs["class"] ==  ['x-p', 'dx-p']:
            c_print("#[#757575]" + "  • " + i.text, end="")
        elif has_class and i.attrs["class"] == ['x-lab', 'dx-lab']:
            c_print("#[#757575]" + " [" + i.text + "]", end="")
        elif has_class and i.attrs["class"] == ['x-pos', 'dx-pos']:
            print(" " + i.text, end="")
        elif parent_has_class and parent.attrs["class"] == ['x-h', 'dx-h']:
            continue
        elif parent_has_class and parent.attrs["class"] == ['x-p', 'dx-p']:
            continue
        elif parent_has_class and parent.attrs["class"] == ['x-lab', 'dx-lab']:
            continue
        else:
            c_print("#[#757575]" + " " + i.text, end="")
    print()


def print_synonym(block):
    s_title = block.strong.text.upper()
    c_print("#[bold #757575]" + "\n" + s_title)

    for item in block.find_all("div", ["item lc lc1 lpb-10 lpr-10", "item lc lc1 lc-xs6-12 lpb-10 lpr-10"]):
        print_tag(item)


def parse_synonym(block):
    s_block = block.find("div", re.compile("xref synonyms? hax dxref-w( lmt-25)?"))
    if s_block is not None:
        print_synonym(s_block)
    elif "synonym" in block.attrs["class"]:
        print_synonym(block)


def parse_see_also_lmb(def_block):
    see_also_block = def_block.find("div", "xref see_also hax dxref-w lmt-25 lmb-25")
    if see_also_block is not None:
        see_also = see_also_block.strong.text.upper()
        c_print("#[bold #757575]" + "\n" + see_also)
        for item in see_also_block.find_all("div", ["item lc lc1 lpb-10 lpr-10", "item lc lc1 lc-xs6-12 lpb-10 lpr-10"]):
            print_tag(item)


def parse_see_also(def_block):
    see_also_block = def_block.find("div", re.compile("xref see_also hax dxref-w( lmt-25)?"))
    if see_also_block is not None:
        see_also = see_also_block.strong.text.upper()
        c_print("#[bold #757575]" + "\n" + see_also)
        for item in see_also_block.find_all("div", ["item lc lc1 lpb-10 lpr-10", "item lc lc1 lc-xs6-12 lpb-10 lpr-10"]):
            print_tag(item)


def parse_compare(def_block):
    compare_block = def_block.find("div", re.compile("xref compare hax dxref-w( lmt-25)?"))

    if compare_block is not None:
        compare = compare_block.strong.text.upper()
        c_print("#[bold #757575]" + "\n" + compare)
        for item in compare_block.find_all("div", ["item lc lc1 lpb-10 lpr-10", "item lc lc1 lc-xs6-12 lpb-10 lpr-10"]):
            print_tag(item)


def parse_usage_note(def_block):
    usage_block = def_block.find("div", "usagenote dusagenote daccord")

    if usage_block is not None:
        usagenote = usage_block.h5.text
        c_print("#[bold #757575]" + "\n  " + usagenote)
        for item in usage_block.find_all("li", "text"):
            item = item.text
            c_print("#[#757575]" + "    " + item)


def parse_def(def_block):
    if "phrase-body" in def_block.parent.attrs["class"]:
        parse_meaning(def_block, True)
        parse_example(def_block, True)
    else:
        parse_meaning(def_block)
        parse_example(def_block)

    parse_synonym(def_block)
    parse_see_also(def_block)
    parse_compare(def_block)
    parse_usage_note(def_block)

def parse_idiom(block):
    idiom_block = block.find("div", re.compile("xref idioms? hax dxref-w lmt-25 lmb-25"))

    if idiom_block is not None:
        idiom_title = idiom_block.h3.text.upper()
        c_print("#[bold #757575]" + "\n" + idiom_title)
        for item in idiom_block.find_all("div", ["item lc lc1 lpb-10 lpr-10", "item lc lc1 lc-xs6-12 lpb-10 lpr-10"]):
            print_tag(item)


def parse_sole_idiom(block):
    idiom_sole_meaning = block.find("div", "def ddef_d db")

    text = idiom_sole_meaning.text.strip()
    if text[-1] == ":":
        text = text[ : -1]

    if idiom_sole_meaning is not None:
        print("\033[34;1m: \033[0m" + "\033[34m" + text + "\033[0m")

    parse_example(block)
    parse_see_also(block)


def parse_phrasal_verb(block):
    pv_block = block.find("div", re.compile("xref phrasal_verbs? hax dxref-w lmt-25 lmb-25"))

    if pv_block is not None:
        pv_title = pv_block.h3.text.upper()
        c_print("#[bold #757575]" + "\n" + pv_title)
        for item in pv_block.find_all("div", ["item lc lc1 lc-xs6-12 lpb-10 lpr-10", "item lc lc1 lpb-10 lpr-10"]):
            print_tag(item)


def parse_dict_body(block):
    #TODO https://dictionary.cambridge.org/dictionary/english/best-efforts  parse <div class="pr runon drunon">
    subblocks = block.find_all("div", ["pr dsense", "pr dsense dsense-noh"])

    if len(subblocks) != 0:
        for subblock in subblocks:
            sense = subblock.find("div", "sense-body dsense_b")
            if sense is not None:
                for child in sense.children:
                    try:
                        attr = child.attrs["class"]
                        if attr and attr == ["def-block", "ddef_block"]:
                            parse_def(child)

                        elif attr and "synonym" in attr:
                            parse_synonym(child)

                        elif attr and (attr == ["pr", "phrase-block", "dphrase-block", "lmb-25"] or attr == ["pr", "phrase-block", "dphrase-block"]):
                            parse_ptitle(child)

                            for i in child.find_all("div", "def-block ddef_block"):
                                parse_def(i)
                    except Exception:
                        pass

    else:
        idiom_sole_block = block.find("div", "idiom-block")
        if idiom_sole_block is not None:
            parse_sole_idiom(idiom_sole_block)

        parse_compare(block)

    parse_see_also_lmb(block)
    parse_idiom(block)
    parse_phrasal_verb(block)


def parse_dict_name(first_dict):
    if first_dict.small is not None:
        dict_info = replace_all(first_dict.small.text).strip("(").strip(")")
        dict_name = dict_info.split("©")[0]
        dict_name = dict_name.split("the")[-1]
    else:
        dict_name="Cambridge Dictionary"
    c_print(f"#[#757575]{dict_name}", justify="right")



"""
cambridge is a terminal version of Cambridge Dictionary.
The dictionary data comes from https://dictionary.cambridge.org
If you're not satisfied with the result, you can add "-w" flag to look up the word in Merriam-Webster Dictionary.
"""

__version__ = "3.13.11"



import sys
import asyncio
from lxml import etree # type: ignore

from .console import c_print
from .utils import fetch, get_request_url, decode_url, OP, DICT, has_tool, get_suggestion, get_suggestion_by_fzf, get_wod_selection, get_wod_selection_by_fzf, quit_on_no_result, cancel_on_error, cancel_on_error_without_retry
from .log import logger
from .cache import check_cache, save_to_cache, get_cache
from . import camb
from . import color as w_col

WEBSTER_BASE_URL = "https://www.merriam-webster.com"
WEBSTER_DICT_BASE_URL = WEBSTER_BASE_URL + "/dictionary/"
WEBSTER_WORD_OF_THE_DAY_URL = WEBSTER_BASE_URL + "/word-of-the-day"
WEBSTER_WORD_OF_THE_DAY_URL_CALENDAR = WEBSTER_BASE_URL + "/word-of-the-day/calendar"

search_pattern = """
//*[@id="left-content"]/div[contains(@id, "-entry")] |
//*[@id="left-content"]/div[@id="phrases"] |
//*[@id="left-content"]/div[@id="synonyms"] |
//*[@id="left-content"]/div[@id="examples"]/div[@class="content-section-body"]/div[contains(@class,"on-web-container")]/div[contains(@class,"on-web")] |
//*[@id="left-content"]/div[@id="related-phrases"] |
//*[@id="left-content"]/div[@id="nearby-entries"]
"""
parser = etree.HTMLParser(remove_comments=True)

word_entries = []    # A page may have multiple word entries, e.g. "runaway" as noun, "runaway" as adjective, "run away" as verb
word_forms = set()   # A word may have multiple word forms, e.g. "ran", "running", "run", "flies"
word_types = set()   # A word's word types, e.g. "preposition", "adjective"


async def search_webster(session, input_word, is_fresh=False, no_suggestions=False, req_url=None):
    if req_url is None:
        req_url = get_request_url(WEBSTER_DICT_BASE_URL, input_word, DICT.MERRIAM_WEBSTER.name)

    if is_fresh:
        await fresh_run(session, input_word, no_suggestions, req_url)
    else:
        res_url = await check_cache(input_word, req_url)
        if res_url is None:
            logger.debug(f'{OP.NOT_FOUND.name} "{input_word}" in cache')
            await fresh_run(session, input_word, no_suggestions, req_url)
        elif DICT.CAMBRIDGE.name.lower() in res_url:
            await camb.cache_run(res_url)
        else:
            await cache_run(res_url)


async def cache_run(res_url_from_cache):
    res_word, res_text = await get_cache(res_url_from_cache)
    logger.debug(f'{OP.FOUND.name} "{res_word}" from {DICT.MERRIAM_WEBSTER.name} in cache')
    logger.debug(f"{OP.PARSING.name} {res_url_from_cache}")
    tree = etree.HTML(res_text, parser)
    nodes = tree.xpath(search_pattern)
    await parse_and_print(nodes, res_url_from_cache, new_line=False)
    c_print(f'\n#[#757575]{OP.FOUND.name} "{res_word}" from {DICT.MERRIAM_WEBSTER.name} in cache. You can add "-f" to fetch the {DICT.CAMBRIDGE.name} dictionary')


async def fresh_run(session, input_word, no_suggestions, req_url):
    response = await fetch(session, req_url)
    res_url = str(response.real_url)
    status = response.status

    # By default Requests will perform location redirection for all verbs except HEAD.
    # https://requests.readthedocs.io/en/latest/user/quickstart/#redirection-and-history
    # You don't need to deal with redirection yourself.
    # if status == 301:
    #     loc = res.headers["location"]
    #     new_url = WEBSTER_BASE_URL + loc
    #     new_res = await fetch(session, new_url)

    res_text = None
    attempt = 0
    while True:
        try:
            res_text = await response.text()
        except asyncio.TimeoutError as error:
            attempt = cancel_on_error(req_url, error, attempt, OP.FETCHING.name)
            continue
        # There is also a scenario that in the process of cancelling, while is still looping, leading to run coroutine with a closed session.
        # If session is closed, and you go on connecting, ClientConnectionError will be throwed.
        except Exception as error:
            cancel_on_error_without_retry(req_url, error, OP.FETCHING.name)
            break
        else:
            break

    if res_text is not None:
        if status == 200:
            logger.debug(f'{OP.FOUND.name} "{input_word}" in {DICT.MERRIAM_WEBSTER.name} at {res_url}')

            logger.debug(f"{OP.PARSING.name} {res_url}")
            tree = etree.HTML(res_text, parser)

            partial_match = tree.xpath('//p[contains(@class,"partial")]')
            if partial_match:
                input_word = decode_url(res_url).split("/")[-1]
                suggestions = tree.xpath('//h2[@class="hword"]/text() | //h2[@class="hword"]/span/text()')

                logger.debug(f"{OP.PRINTING.name} out suggestions at {res_url}")
                select_word = get_suggestion_by_fzf(suggestions, DICT.MERRIAM_WEBSTER.name) if has_tool("fzf") else get_suggestion(suggestions, DICT.MERRIAM_WEBSTER.name)
                if select_word == "":
                    logger.debug(f'{OP.SWITCHED.name} to {DICT.CAMBRIDGE.name}')
                    await camb.search_cambridge(session, input_word, True, False, no_suggestions, None)
                else:
                    logger.debug(f'{OP.SELECTED.name} "{select_word}"')
                    await search_webster(session, select_word, False, no_suggestions, None)
            else:
                sub_tree = tree.xpath('//*[@id="left-content"]')[0]
                nodes = sub_tree.xpath(search_pattern)
                if len(nodes) == 0:
                    quit_on_no_result(DICT.MERRIAM_WEBSTER.name, is_spellcheck=False)

                # Response word within res_url is not same with what apppears on the web page. e.g. "set in stone"
                result = sub_tree.xpath('//*[@id="left-content"]/div[contains(@id, "-entry-1")]/div[1]/div/div[1]/h1/text()') \
                       or sub_tree.xpath('//*[@id="left-content"]/div[contains(@id, "-entry-1")]/div[1]/div/div/h1/span/text()')

                if len(result) == 0:
                    quit_on_no_result(DICT.MERRIAM_WEBSTER.name, is_spellcheck=False)

                await parse_and_print(nodes, res_url, new_line=True)

                sub_text = etree.tostring(sub_tree).decode('utf-8')
                # logger.debug(f'START CACHING: input_word is "{input_word}"; res_word is "{res_word}"; res_url is "{res_url}"')
                await save_to_cache(input_word, result[0], res_url, sub_text)

        elif status == 404:
            logger.debug(f'{OP.NOT_FOUND.name} "{input_word}" at {res_url}')

            if no_suggestions:
                sys.exit(-1)

            logger.debug(f"{OP.PARSING.name} out suggestions at {res_url}")
            tree = etree.HTML(res_text, parser)
            result = tree.xpath('//div[@class="widget spelling-suggestion"]')
            if len(result) == 0:
                quit_on_no_result(DICT.MERRIAM_WEBSTER.name, is_spellcheck=True)

            nodes = result[0]
            suggestions = []
            for node in nodes:
                if node.tag != "h1":
                    for word in node.itertext():
                        w = word.strip("\n").strip()
                        if w.startswith("The"):
                            continue
                        else:
                            suggestions.append(w)

            logger.debug(f"{OP.PRINTING.name} out suggestions at {res_url}")
            select_word = get_suggestion_by_fzf(suggestions, DICT.MERRIAM_WEBSTER.name) if has_tool("fzf") else get_suggestion(suggestions, DICT.MERRIAM_WEBSTER.name)
            if select_word == "":
                logger.debug(f'{OP.SWITCHED.name} to {DICT.CAMBRIDGE.name}')
                await camb.search_cambridge(session, input_word, True, False, no_suggestions, None)
            else:
                logger.debug(f'{OP.SELECTED.name} "{select_word}"')
                await search_webster(session, select_word, False, no_suggestions, None)

        else:
            print(f'Something went wrong when fetching {req_url} with STATUS: {status}')
            sys.exit(2)


def nearby_entries(node):
    print()

    for elm in node.iterdescendants():
        try:
            has_title = (elm.tag == "h2")
            has_word = (elm.tag == "div" or elm.tag == "span") and (elm.attrib["class"] == "b-link hw-text fw-bold")
            has_nearby = (elm.tag == "a") and (elm.attrib["class"] == "b-link")
            has_em = (elm.tag == "em")
        except KeyError:
            continue
        else:
            if has_title and "Browse Nearby" in elm.text:
                c_print(f"\n#[bold {w_col.nearby_title}]{elm.text}", end="\n")
            elif has_title and "Dictionary Entries Near" in elm.text:
                c_print(f"\n#[bold {w_col.nearby_title}]{elm.text}", end="")
            elif has_em:
                word = "".join(list(elm.itertext()))
                c_print(f"#[bold {w_col.nearby_em}]{word}", end="\n")
            elif has_word:
                c_print(f"#[{w_col.nearby_word}]{elm.text}", end="\n")
            elif has_nearby:
                c_print(f"#[{w_col.nearby_item}]{elm.text}", end="\n")

def synonyms(node):
    print()

    for elm in node.iterdescendants():
            if elm.tag == "h2":
                c_print(f"#[bold {w_col.syn_title}]{elm.text}", end="\n")

            if elm.tag == "p" and elm.attrib["class"] == "function-label":
                c_print(f"#[{w_col.syn_label}]{elm.text}")

            if elm.tag == "ul":
                children = elm.getchildren()
                total_num = len(children)

                for index, child in enumerate(children):
                    syn = "".join(list(child.itertext())).strip()
                    if index != (total_num - 1):
                        c_print(f"#[{w_col.syn_item}]{syn},", end=" ")
                    else:
                        c_print(f"#[{w_col.syn_item}]{syn}", end="\n")


def examples(node):
    print()

    for elm in node.iterdescendants():
        try:
            is_title = ("ex-header function-label" in elm.attrib["class"])
            has_aq = (elm.attrib["class"] == "t has-aq")
        except KeyError:
            continue
        else:
            if is_title:
                c_print(f"#[{w_col.eg_title} bold]{elm.text}", end="")
            if has_aq:
                texts = list(elm.itertext())

                for index, t in enumerate(texts):
                    if index == 0:
                        c_print(f"\n#[{w_col.accessory}]|", end="")
                        c_print(f"#[{w_col.eg_sentence}]{t}", end="")
                    else:
                        hit = False
                        text = t.strip().lower()
                        for w in word_entries:
                            if "preposition" in word_types or "adverb" in word_types or "conjuction" in word_types and ("noun" in word_types and text[-1] !="s"):
                                if w == text:
                                    hit = True
                                    break
                            else:
                                if w in text and len(text) < 20:
                                    hit = True
                                    break
                                elif '/' in w:
                                    temp = w.split("/")
                                    if temp[0] in text or temp[-1] in text:
                                        hit = True
                                        break
                                    break

                        for f in word_forms:
                            if f == text:
                                hit = True
                                break

                        if hit:
                            c_print(f"#[{w_col.eg_word} bold]{t}", end="")
                        else:
                            c_print(f"#[{w_col.eg_sentence}]{t}", end="")


def phrases(node):
    print()
    children = node.getchildren()[1]
    for child in children:
        try:
            if child.attrib["class"] == "drp":
                if child.getnext().tag == "span":
                    c_print(f"#[{w_col.ph_item} bold]{child.text}", end="")
                else:
                    c_print(f"#[{w_col.ph_item} bold]{child.text}", end="\n")

            elif child.attrib["class"] == "vg":
                vg(child)

        except KeyError:
            for i in child.getchildren():
                if i.attrib["class"] == "vl":
                    print_or_badge(i.text)
                else:
                    c_print(f"#[{w_col.ph_item} bold]{i.text}", end="")
            if child.getnext().get("class") == "vg":
                print()


def related_phrases(node):
    print()

    children = node.getchildren()

    title = children[1]
    texts = list(title.itertext())

    for t in texts:
        if t.strip():
            if t.lower() in word_entries:
                c_print(f"#[{w_col.rph_em} bold]{t}", end="\n")
            else:
                c_print(f"#[{w_col.rph_title} bold]{t}", end="")

    pr_sec = children[2]
    phrases = []
    for i in pr_sec.iterdescendants():
        if i.tag == "li" and "related-phrases-list-item" in i.get("class"):
            ts = "". join(list(i.itertext())).strip("\n").strip()
            if ts not in phrases:
                phrases.append(ts)
    for index, phrase in enumerate(phrases):
        text = phrase + ", " if index != len(phrases) -1  else phrase
        c_print(f"#[{w_col.rph_item}]{text}", end="")


def dtText(node, ancestor_attr, num_label_count):
    """Print the meaning text starting with `:`. E.g. one of the word replenish's meaning texts `: to fill or build up again`"""
    texts = list(node.itertext())

    l_words = get_word_cases(node)[0]
    u_words = get_word_cases(node)[1]

    node_pre = node.getprevious()
    node_pre_attr = None
    if node_pre is not None:
        node_pre_attr = node_pre.get("class")

    prefix = "\n " if num_label_count == 2 else "\n"

    is_format = False
    if node_pre_attr == "sub-content-thread" or node_pre_attr == "uns" or node_pre_attr == "dt ":
        format_basedon_ancestor(ancestor_attr, prefix=prefix)
        is_format = True

    for index, text in enumerate(texts):
        if text.strip("\n").strip():
            if text == ": ":
                print_meaning_content(text, end="")
            elif text in [" see also ", " see ", " compare "]: # e.g. "drag", "group"
                if is_format:
                    print_meaning_keyword(text.strip().upper())
                elif "has-sn" in ancestor_attr or "has-num" in ancestor_attr:
                    format_basedon_ancestor(ancestor_attr, prefix=prefix)
                    print_meaning_keyword(text.strip().upper())
                else:
                    print_meaning_keyword("\n" + text.strip().upper())
            elif u_words and text in u_words:
                text_new = text.upper()
                print_meaning_content(text_new, end="")
            elif l_words and text in l_words:
                text_new = (" " + text)
                print_meaning_content(text_new, end="")
            #elif index == len(texts) - 1:
            #    text = text[:-1] if text[-1] == " " else text
            #    print("$$$$$$")
            #    print_meaning_content(text, end="") # e.g. creative
            else:
                print_meaning_content(text, end="")


def print_mw(text, nospace, tag):
    end = "" if nospace else " "
    bold = "" if tag == "normal" else "bold"
    c_print(f"#[{w_col.meaning_sentence}{bold}]{text}", end=end)


def get_word_cases(node):
    l_words = []
    u_words = []
    for i in node.iterdescendants():
        attr = i.get("class")
        if attr is not None:
            if "lowercase" in attr:
                l_words.append(i.text)
            elif "uppercase" in attr:
                u_words.append(i.text)
    return l_words, u_words


def get_word_faces(node):
    hl_words = []
    ems = []
    for i in node.iterdescendants():
        attr = i.get("class")
        if attr is not None:
            if i.tag == "em" and "mw" in attr:
                ems.append(i.text)
            elif i.tag == "span" and "mw" in attr and "mw_t_sp" != attr and "mw_t_gloss" != attr:
                hl_words.append(i.text)
    return hl_words, ems


def ex_sent(node, ancestor_attr, num_label_count):
    if ancestor_attr:
        format_basedon_ancestor(ancestor_attr, prefix="\n")
    else:
        print()

    if num_label_count == 2:
        print(" ", end="")

    c_print(f"#[{w_col.accessory}]|", end="")

    hl_words = get_word_faces(node)[0]
    ems = get_word_faces(node)[1]

    texts = list(node.itertext())
    count = len(texts)

    for index, t in enumerate(texts):
        text = t.strip("\n").strip()
        if text:
            if t in hl_words:
                if index == count - 1 or (index == count - 2 and (texts[count-1].strip("\n").strip() == "")):
                    print_mw(text, True, "hl")
                elif texts[index + 1][0].isalpha() or (texts[index + 1].strip("\n").strip() and texts[index + 1].strip("\n").strip()[0] in [",", ".", "?", "!", "-", "—"]):
                    print_mw(text, True, "hl")
                elif index != count - 2  and not texts[index + 1].strip("\n").strip() and texts[index + 2].strip("\n").strip() in ems:
                    print_mw(text, True, "hl")
                else:
                    print_mw(text, False, "hl")
            elif t in ems:
                if index != 0 and texts[index - 1].endswith(" "):
                    print("", end = " ")
                c_print(f"#[{w_col.meaning_sentence} bold]{text}", end = "")
                if index != (count - 1) and texts[index + 1] != " " and texts[index + 1].startswith(" ") :
                    print("", end = " ")
            else:
                if index == count - 1 or (index == count - 2 and texts[count-1].strip("\n").strip() == "") :
                    print_mw(text, True, "normal")
                elif texts[index + 1] in ems:
                    print_mw(text, True, "normal")
                elif texts[index + 1] in hl_words and (t[-1] == '"' or t[-1] == '-'):
                    print_mw(text, True, "normal")
                else:
                    print_mw(text, False, "normal")


def sub_content_thread(node, ancestor_attr, num_label_count=1):
    children = node.getchildren()
    for child in children:
        attr = child.attrib["class"]

        if ("ex-sent" in attr) and ("aq has-aq" not in attr):
            ex_sent(child, ancestor_attr, num_label_count)

        elif "vis" in attr:
            elms = child.getchildren()
            for e in elms:
                elm = e.getchildren()[0]
                elm_attr = elm.attrib["class"]
                if ("ex-sent" in elm_attr) and ("aq has-aq" not in elm_attr):
                    ex_sent(elm, ancestor_attr, num_label_count)


def extra(node, ancestor_attr):
    texts = list(node.itertext())

    l_words = get_word_cases(node)[0]
    u_words = get_word_cases(node)[1]

    prev = node.getprevious()
    prev_attr = ""
    if prev is not None:
        prev_attr = prev.get("class")

    for text in texts:
        text_new = text.strip("\n").strip()
        if text_new:
            if text_new == "called also" or text_new == "compare":
                prefix = "\n" if prev_attr == "sub-content-thread" else " "  # e.g. "flash-bang"
                print_meaning_keyword(prefix + text_new.upper())
            elif u_words and text in u_words:
                text_new = text_new.upper()
                print_meaning_content(text_new, end="")
            elif l_words and text in l_words:
                text_new = (" " + text_new)
                print_meaning_content(text_new, end="")
            elif text_new == ",":
                print_meaning_content(text_new, end=" ")
            else:
                print_meaning_content(text_new, end="")

    print("", end = " ")


def unText_simple(node, ancestor_attr, has_badge=True):
    if not has_badge:
        print()
        format_basedon_ancestor(ancestor_attr, prefix="")

    prefix = f"#[{w_col.meaning_arrow}]->"
    suffix = " "

    node_pre = node.getprevious()
    if node_pre is not None and node_pre.get("class") == "mdash mdash-silent":
        node_p_pre = node.getparent().getprevious()
        if node_p_pre is not None and node_p_pre.get("class") == "un":
            c_print(" " + prefix, end=suffix)
        else:
            c_print(prefix, end=suffix)

    bolds = get_word_faces(node)
    text_list = list(node.itertext())

    for bold in bolds:
        if bold:
            for index, t in enumerate(text_list):
                if t in bold:
                    #TODO "/" should only reset the current effect (bold), not including w_col.meaning_arrow
                    text_list[index] = f"#[bold]{t}#[/bold]#[{w_col.meaning_arrow}]"

    text = "".join(text_list).strip()
    c_print(f"#[{w_col.meaning_arrow}]{text}", end="")


def sense(node, attr, parent_attr, ancestor_attr, num_label_count):
    """e.g. sense(node, "sense has-sn", "sb-0 sb-entry, "sb has-num has-let ms-lg-4 ms-3 w-100", 1)"""

    sense_content = None
    children = node.getchildren()

    # meaning without any sign
    if attr == "sense  no-subnum":
        sense_content = children[0] # class "sense-content w-100"

    # meaning with "1" + "a"
    elif attr == "sense has-sn has-num":
        sn = children[0].getchildren()[0].text

        node_prev = node.getprevious()
        if "has-subnum" in ancestor_attr and node_prev is None and "sb-0" in parent_attr:
            c_print(f"#[bold {w_col.meaning_letter}]{sn}", end = " ")
        elif "has-subnum" in ancestor_attr and (node_prev is not None or parent_attr != "pseq no-subnum"):
            if num_label_count == 2:
                print(" ", end="")
            c_print(f"  #[bold {w_col.meaning_letter}]{sn}", end = " ")
        else:
            c_print(f"#[bold {w_col.meaning_letter}]{sn}", end = " ")

        sense_content = children[1] # class "sense-content w-100"

    # meaing with only "b" or "1" + "a" + "(1)", or "1" + "a"
    elif attr == "sense has-sn" or attr == "sen has-sn":
        sn = children[0].getchildren()[0].text

        if "has-subnum" in ancestor_attr and "sb-0" in parent_attr:
            c_print(f"#[bold {w_col.meaning_letter}]{sn}", end = " ")
        else:
            if "letter-only" in ancestor_attr:
                if "sb-0" not in parent_attr:
                    print("  ", end="")
                c_print(f"#[bold {w_col.meaning_letter}]{sn}", end = " ")
            else:
                if num_label_count == 2 and sn == "a":
                    if "sb-0 sb-entry" != ancestor_attr:
                        c_print(f"   #[bold {w_col.meaning_letter}]{sn}", end = " ")
                    else:
                        c_print(f"#[bold {w_col.meaning_letter}]{sn}", end = " ")
                elif num_label_count == 2:
                    c_print(f"   #[bold {w_col.meaning_letter}]{sn}", end = " ")
                else:
                    c_print(f"  #[bold {w_col.meaning_letter}]{sn}", end = " ")

        if node.tag == "span": # e.g. "knife and fork, track intransitive verb 2"
            # Intentionally take out this part from tags(), don't merge in later.
            # Because using tags() here conflicts with these same classes elsewhere with different new line or not requirements.
            # Adding extra checks as to where the node comes from in already complex tags() is not a good idea.
            for child in children[1 : ]:
                child_attr = child.get("class")
                if child_attr is not None:
                    if child_attr == "if":
                        end = "\n" if child.getnext() is None else ""
                        c_print(f"#[bold]{child.text.strip()}", end=end)
                    elif "badge mw-badge-gray-100" in child_attr:
                        end = " " if child.getnext() is not None else "\n"
                        print_meaning_badge(child.text, end=end)
                else:
                    for c in child:
                        c_attr = c.get("class")
                        if c_attr == "vl":
                            print_meaning_badge(c.text.strip(), end=" ")
                        elif c_attr == "va":
                            print_class_va(c.text)
                        elif "prons-entries-list" in c_attr:
                            print_pron(c)
                            print()
        else:
            sense_content = children[1]

    # meaning with only (2)
    elif attr == "sense has-num-only has-subnum-only":
        if num_label_count == 2:
            print(" ", end="")
        if "letter-only" in ancestor_attr:
            if children[0].attrib["class"] == "sn":
                print("    ", end = "")
            else:
                print("  ", end="")
        else:
            print("    ", end = "")
        sense_content = children[1] # class "sense-content w-100"

    # meaning with only number
    elif attr == "sense has-sn has-num-only":
        sense_content = children[1] # class "sense-content w-100"

    else:
        sense_content = children[1]

    # "sense-content w-100"
    if sense_content is not None:
        tags(sense_content, attr, num_label_count)


def sb_entry(node, parent_attr, num_label_count):
    child = node.getchildren()[0]
    attr = node.attrib["class"]         # "sb-0 sb-entry"
    child_attr = child.attrib["class"]  # "sense has-sn" or "pseq no-subnum" or "sen has-sn"
    if "pseq" in child_attr:
        elms = child.getchildren()[0].getchildren()
        for e in elms:
            e_attr = e.attrib["class"]  # "sense has-sn"
            sense(e, e_attr, attr, parent_attr, num_label_count)
    elif "sense" in child_attr and child.tag != "span":
        sense(child, child_attr, attr, parent_attr, num_label_count) # e.g. sense(child, "sense has-sn", "sb-0 sb-entry, "sb has-num has-let ms-lg-4 ms-3 w-100", 1)
    elif "sen" in child_attr and child.tag == "span": # e.g. "knife and fork"
        sense(child, child_attr, attr, parent_attr, num_label_count)


def tags(node, ancestor_attr, num_label_count):
    has_badge = True
    has_et = False
    has_dtText = False
    no_print = False
    has_sense2 = False
    sense2_has_blank_span = False

    for elm in node.iterdescendants():
        elm_attr = elm.get("class")
        parent = elm.getparent()
        parent_attr = parent.get("class")

        if elm_attr is not None:
            if "sn sense-2" == elm_attr:
                has_sense2 = True
                if elm.getnext() is not None and elm.getnext().get("class") is None:
                    sense2_has_blank_span = True
            if "badge " in elm_attr and "pron" not in elm_attr:
                prev = elm.getprevious()
                if prev is not None and prev.get("class") is None:
                    print("", end=" ")
                text = "".join(list(elm.itertext())).strip()
                if elm.getnext() is not None:
                    print_meaning_badge(text, end=" ")
                elif "spl plural badge" in elm_attr or "lb badge" in elm_attr or "sl badge" in elm_attr:
                    end = "\n" if has_sense2 else ""
                    print_meaning_badge(text, end=end)
                else:
                    print_meaning_badge(text, end="")

            elif elm_attr == "et":
                et(elm)
                has_et = True

            elif elm_attr == "il ":
                print_meaning_badge(elm.text.strip(), end=" ")

            elif elm_attr == "if":
                print_class_if(elm.text)

            elif elm_attr == "sgram":
                print_class_sgram(elm)

            elif elm_attr == "vl":
                parent_prev = parent.getprevious()
                if parent_prev is None and elm.text[0] == " ":
                    print_meaning_badge(elm.text[1 : ])
                else:
                    print_meaning_badge(elm.text)

            elif elm_attr == "va":
                print_class_va(elm.text.strip())

            elif elm_attr == "sd":
                parent_prev = parent.getprevious()
                if parent_prev is not None and "hasSdSense" in parent_prev.get("class"):
                    print()
                if parent_attr is not None and parent_attr == "sdsense":
                    format_basedon_ancestor(ancestor_attr, prefix="")

                if num_label_count == 2:
                    print(" ", end="")

                print_meaning_badge(elm.text, end=" ")

            elif elm_attr == "dtText":
                dtText(elm, ancestor_attr, num_label_count) # only meaning text
                has_dtText = True

            elif "sub-content-thread" in elm_attr:
                sub_content_thread(elm, ancestor_attr, num_label_count) # example under the meaning
                has_badge = False

            elif elm_attr == "ca":
                extra(elm, ancestor_attr)

            elif elm_attr == "unText":
                unText_simple(elm, ancestor_attr, has_badge)

            elif "prons-entries-list" in elm_attr and parent_attr is None:
                print_pron(elm)

            elif elm_attr == "sn sense-2":
                no_print = True
    if (not has_et and not no_print) or (has_et and has_dtText) or (has_sense2 and sense2_has_blank_span):
        print()


def vg_sseq_entry_item(node):
    """Print one meaning of one entry(noun entry, adjective entry, or verb entry and so forth). e.g. 1: the monetary worth of something."""

    num_label_count = 1
    children = node.getchildren()
    for child in children:
        attr = child.attrib["class"]
        # print number label if any
        if attr == "vg-sseq-entry-item-label":
            c_print(f"#[bold {w_col.meaning_num}]{child.text}", end=" ")
            num_label_count = len(child.text) # important! making sure two-digit numbering and things under it can vertically align properly.

        # print meaning content
        elif "ms-lg-4 ms-3 w-100" in attr:
            for c in child.iterchildren(): # c:  class="sb-0 sb-entry"
                cc = c.getchildren()[0]    # cc: class="sen has-num-only"
                cc_attr = cc.get("class")
                if cc_attr is not None and cc_attr == "sen has-num-only":
                    tags(cc, cc_attr, num_label_count)
                    continue # !!! very important, or sb_entry() will run the node "sb-0 sb-entry" containing "sen has-num-only" once again

                # print class "sb-0 sb-entry", "sb-1 sb-entry" ...
                sb_entry(c, attr, num_label_count)


def et(node):
    text = "".join(node.itertext()).strip("\n")
    if node.getnext() is not None:
        print(text, end=" ")
    else:
        print(text, end="\n")


def vg(node):
    """Print one entry(e.g. 1 of 3)'s all meanings. e.g. 1 :the monetary worth of something 2 :a fair return... 3 :..."""

    children = node.getchildren()
    for child in children:
        attr = child.get("class")
        # print one meaning of one entry
        if attr is not None and "vg-sseq-entry-item" in attr:
            vg_sseq_entry_item(child)

        # print transitive or intransitive
        elif attr is not None and (attr == "vd firstVd" or attr == "vd"):
            e = child.getchildren()[0]
            c_print(f"#[bold]{e.text}")

        # print tags like "informal" and the tags at the same livel with transitives
        elif attr is not None and "sls" in attr:
            e = child.getchildren()[0]
            e_attr = e.get("class")
            if e_attr is not None and "badge" in e_attr:
                print_meaning_badge(e.text, end="")
            else:
                c_print(f"#[bold]{e.text}")

            if "vg-sseq-entry-item" in child.getnext().get("class"):
                print()


def entry_header_content(node):
    """Print entry header content. e.g. value 1 of 3 noun"""

    for elm in node.iterchildren():
        if elm.tag == "h1" or elm.tag == "p":
            word = "".join(list(elm.itertext()))
            global word_entries
            word_entries.append(word.strip().lower())
            end = "" if elm.getnext() is None else " "
            c_print(f"#[{w_col.eh_h1_word} bold]{word}", end=end)

        elif elm.tag == "span":
            num = " ".join(list(elm.itertext()))
            end = "" if elm.getnext() is None else " "
            print(num, end=end)

        elif elm.tag == "h2":
            type = " ".join(list(elm.itertext()))
            c_print(f"#[bold {w_col.eh_word_type}]{type}", end="")
            global word_types
            word_types.add(type.strip().lower())

    print()


def entry_attr(node):
    """Print the pronounciation. e.g. val·ue |ˈval-(ˌ)yü|"""

    for elm in node.iterchildren():
        if "col word-syllables-prons-header-content" in elm.attrib["class"]:
            for i in elm.iterchildren():
                if i.tag == "span" and i.attrib["class"] == "word-syllables-entry":
                    syllables = i.text
                    print(f"{syllables}", end=" ")

                elif i.tag == "span" and "prons-entries-list-inline" in i.attrib["class"]:
                    print_pron(i, True)


def row_entry_header(node):
    """Print class row entry-header, the parent and caller of entry_header_content() and entry_attr()."""

    for elm in node.iterchildren():
        if elm.attrib["class"] == "col-12":
            for i in elm.iterchildren():
                if "entry-header-content" in i.attrib["class"]:
                    entry_header_content(i)
                elif "row entry-attr" in i.attrib["class"]:
                    entry_attr(i)


def entry_uros(node):
    """Print other word forms. e.g. valueless, valuelessness"""

    for elm in node.iterdescendants():
        attr = elm.get("class")
        if attr is not None:
            if elm.tag == "span" and "fw-bold ure" in attr:
                c_print(f"#[bold {w_col.wf}]{elm.text}", end = " ")

            elif elm.tag == "span" and "fw-bold fl" in attr:
                elm_next = elm.getnext()
                if elm_next is not None and elm_next.get("class") == "ins":
                    end = "\n"
                elif elm_next is not None and "badge" in elm_next.get("class"):
                    end = " "
                else:
                    end = ""
                c_print(f"#[{w_col.eh_word_type}]{elm.text}", end=end)

            elif "ins" in attr:
                print("", end="")
                print_class_ins(elm)

            elif "sl badge" in attr:
                text = "".join(list(elm.itertext())).strip()
                prev = elm.getprevious()
                if prev is not None and prev.get("class") == "vrs":
                    print_meaning_badge(" " + text)
                else:
                    print_meaning_badge(text)

            elif "utxt" in attr:
                for i in elm.iterchildren():
                    sub_attr = i.get("class")
                    if sub_attr is not None and "sub-content-thread" in sub_attr:
                        sub_content_thread(i, "")

            elif "prons-entries-list" in attr:
                print_pron(elm)

            elif "vrs" in attr:
                # can't get css element ::before.content like "variants" in the word "duel"
                child = elm.getchildren()[0]
                for c in child.iterchildren():
                    attr_c = c.get("class")
                    if attr_c == "il " or attr_c == "vl":
                        print_or_badge(c.text)
                    elif attr_c == "va":
                        if c.text is None:
                            for i in child:
                                print_class_va(i.text)
                        else:
                            end = " " if (c.getnext() is not None or elm.getnext() is not None) else ""
                            print_class_va(c.text, end=end)
                    elif "prons-entries-list" in attr_c:
                        continue


def row_headword_row_header_ins(node):
    """Print verb types. e.g. valued; valuing"""

    children = node.getchildren()[0].getchildren()[0]
    if "ins" in children.attrib["class"]:
        print_class_ins(children)
        print()


def print_vrs(node):
    for elm in node.iterchildren():
        elm_attr = elm.get("class")
        if elm_attr is not None and "badge mw-badge-gray-100 text-start text-wrap d-inline" in elm_attr:
            c_print(f"#[bold]{elm.text.strip()}", end="")
        else:
            for child in elm.iterdescendants():
                attr = child.get("class")
                if attr is not None:
                    if attr == "il " or attr == "vl":
                        print_or_badge(child.text)
                    elif attr == "va":
                        end = " " if child.getnext() is not None else ""
                        if child.text is None:
                            for i in child:
                                print_class_va(i.text, end=end)
                        else:
                            print_class_va(child.text, end=end)
                    elif "prons-entries-list" in attr:
                        print_pron(child)
                    else:
                        continue


def row_headword_row_header_vrs(node):
    """Print word variants. e.g. premise variants or less commonly premiss"""

    children = node.getchildren()[0].getchildren()[0] # class "entry-attr vrs"
    print_vrs(children)
    print()


def dxnls(node):
    """Print dxnls section, such as 'see also', 'compare' etc."""

    texts = list(node.itertext())
    for text in texts:
        text = text.strip()
        if not text:
            continue
        if text == "see also":
            c_print(f"#[bold {w_col.dxnls_content}]{text.upper()}", end = " ")
        elif text == "compare":
            c_print(f"#[bold {w_col.dxnls_content}]{text.upper()}", end = " ")
        elif text == ",":
            print_meaning_content(text, end=" ")
        else:
            print_meaning_content(text, end="")

    print()


def dictionary_entry(node):
    """Print one entry of the word and its attributes like plural types, pronounciations, tenses, etc."""

    print()
    for elm in node.iterchildren():
        elm_attr = elm.get("class")
        if elm_attr is not None:
            if "row entry-header" in elm_attr:
                row_entry_header(elm)

            elif elm_attr == "row headword-row header-ins":
                row_headword_row_header_ins(elm)

            elif elm_attr == "row headword-row header-vrs":
                row_headword_row_header_vrs(elm)

            elif elm_attr == "vg":
                vg(elm)

            elif "entry-uros" in elm_attr:
                for i in elm.iterchildren():
                    entry_uros(i)
                    print()

            elif elm_attr == "dxnls":
                dxnls(elm)

            elif elm_attr == "mt-3":
                badge = elm.getchildren()[0] # class "lbs badge mw-badge-gray-100 text-start text-wrap d-inline"
                print_header_badge(badge.text, end="\n")

            elif elm_attr == "cxl-ref":
                texts = list(elm.itertext())
                for text in texts:
                    t = text.strip()
                    if t == ",":
                        print_meaning_content(t, end=" ")
                    elif "chiefly" in t:
                        print_meaning_badge("\n" + t, end=" ")
                    else:
                        print_meaning_content(t.upper(), end="")
                print()


def print_meaning_badge(text, end=""):
    c_print(f"#[{w_col.meaning_badge}]{text}", end=end)


def print_header_badge(text, end=" "):
    c_print(f"#[{w_col.meaning_badge}]{text}", end=end)


def print_meaning_keyword(text, end=" "):
    c_print(f"#[bold {w_col.meaning_keyword}]{text}", end=end)


def print_meaning_content(text, end=""):
    if text == ": ":
        c_print(f"#[{w_col.meaning_content} bold]{text}", end=end)
    else:
        c_print(f"#[{w_col.meaning_content}]{text}", end=end)


def format_basedon_ancestor(ancestor_attr, prefix="", suffix=""):
    print(prefix, end="")
    if ancestor_attr == "sense has-sn has-num-only":
        print("  ", end=suffix)
    elif ancestor_attr == "sense has-sn has-num":
        print("    ", end=suffix)
    elif ancestor_attr == "sense has-sn":
        #if "no-sn letter-only" in root_attr:
        #    print("  ", end=suffix)
        print("    ", end=suffix)
    elif ancestor_attr == "sense  no-subnum":
        print("", end=suffix)
    elif ancestor_attr == "sense has-num-only has-subnum-only":
        print("    ", end=suffix)


def print_pron(node, header=False):
    sibling = node.getnext()
    before_semicolon = ((sibling is not None) and (sibling.get("class") == "sep-semicolon"))
    before_or = ((sibling is not None) and (sibling.get("class") == "il "))

    prons = []
    for text in node.itertext():
        text = text.strip("\n").strip()
        if text and len(text) > 1:
            prons.append(text)

    if "US" in prons[0] or "Canadian" in prons[0] or "British" in prons[0]:
        prons = prons[1:]

    count = len(prons)
    if count == 1:
        if sibling is None:
            if header:
                print(f"|{prons[0]}|", end="\n") # e.g. fortissimo 1 of 2
            else:
                print(f"|{prons[0]}|", end="")   # e.g. fortissimo 2 of 2
        else:
            if before_semicolon or before_or:
                print(f"|{prons[0]}|", end="")
            else:
                print(f"|{prons[0]}|", end=" ")
    if count > 1:
        for index, pron in enumerate(prons):
            if index == 0:
                print(f"|{pron}|", end=" ")
            elif index == count - 1:
                if sibling is not None and sibling.get("class") != "sep-semicolon":
                    c_print(f"#[{w_col.eh_word_syllables}]{pron}", end=" ")
                else:
                    if header:
                        c_print(f"#[{w_col.eh_word_syllables}]{pron}", end="\n")
                    else:
                        c_print(f"#[{w_col.eh_word_syllables}]{pron}", end="")
            elif pron == "," or pron == ";":
                continue
            else:
                text = pron + ", "
                c_print(f"#[{w_col.eh_word_syllables}]{text}", end="")


def print_or_badge(text):
    c_print(f"#[{w_col.or_badge}]{text}", end = "")


def print_class_if(text, before_semicolon=False, before_il=False, has_next_sibling=True):
    if before_semicolon or before_il or not has_next_sibling:
        c_print(f"#[bold]{text}", end="")
    else:
        c_print(f"#[bold]{text}", end=" ")


def print_class_va(text, end= " "):
    c_print(f"#[bold]{text}", end=end)


def print_class_sgram(node):
    for t in node.itertext():
        text = t.strip("\n").strip()
        if text and text.isalpha():
            c_print(f"#[bold]{t}", end=" ")


def print_class_ins(node):
    """print node whose class name includes ins, such as 'ins', 'vg-ins'."""

    for child in node:
        attr = child.get("class")
        if attr is not None:
            if attr == "il  il-badge badge mw-badge-gray-100":
                print_header_badge(child.text.strip(), end=" ") # e.g. "natalism"
            elif attr == "prt-a":
                print_pron(child)
            elif attr == "il ":
                print_or_badge(child.text)
            elif attr == "sep-semicolon":
                print(f"{child.text}", end="")
            elif attr == "if":
                next_sibling = child.getnext()
                if next_sibling is None:
                    print_class_if(child.text, has_next_sibling=False)
                else:
                    sub_attr = next_sibling.get("class")
                    if sub_attr == "sep-semicolon":
                        print_class_if(child.text, before_semicolon=True)
                    elif sub_attr == "il ":
                        print_class_if(child.text, before_il=True)
                    else:
                        print_class_if(child.text, before_semicolon=False)
                global word_forms
                word_forms.add(child.text.strip().lower())
            else:
                c_print(f"#[bold]{child.text}", end="")


def print_dict_name():
    dict_name = "The Merriam-Webster Dictionary"
    c_print(f"#[{w_col.dict_name}]{dict_name}", justify="right")


async def parse_and_print(nodes, res_url, new_line=False):
    logger.debug(f"{OP.PRINTING.name} the parsed result of {res_url}")

    for node in nodes:
        try:
            attr = node.attrib["id"]
        except KeyError:
            attr = node.attrib["class"]

        if "-entry" in attr:
            dictionary_entry(node)

        elif attr == "phrases":
            phrases(node)

        elif attr == "nearby-entries":
            nearby_entries(node)

        elif attr == "synonyms":
            synonyms(node)

        elif "on-web" in attr:
            examples(node)

        elif attr == "related-phrases":
            related_phrases(node)

    if new_line:
        print()


# --- Word of the Day --- #
def print_wod_header(node):
    for elm in node.iterdescendants():
        attr = elm.get("class")
        if attr == "w-a-title":
            for c in elm.iterchildren():
                c_print(f"#[{w_col.wod_title} bold]{c.text}", end="")
            print()

        elif attr == "word-header-txt":
            c_print(f"#[bold]{elm.text}")

        elif attr == "main-attr":
            c_print(f"#[{w_col.wod_type}]{elm.text}", end="")
            print(" | ", end="")

        elif attr == "word-syllables":
            c_print(f"#[{w_col.wod_syllables}]{elm.text}")


def print_wod_p(node):
    text = node.text
    if text:
        print(text, end="")

    for child in node.iterchildren():
        if child is not None and child.tag == "em":
            t = "".join(list(child.itertext()))
            c_print(f"#[bold]{t}", end="")
            print(child.tail, end="")
        elif child is not None and child.tag == "a":
            child_text = child.text
            child_tail = child.tail
            if child_text == "See the entry >":
                continue
            else:
                if child_text is not None:
                    print(child_text, end="")
                for c in child.iterchildren():
                    if c is not None and c.tag == "em":
                        c_print(f"#[bold]{c.text}", end="")

            if child_tail is not None:
                print(child_tail, end="")
    print()


def print_wod_def(node):
    for elm in node.iterchildren():
        tag = elm.tag

        if tag == "h2":
            text = elm.text.strip("\n").strip()
            if text:
                c_print(f"\n#[{w_col.wod_subtitle} bold]{text}")
            children = list(elm.iterchildren())
            if children:
                child = children[0]
                tail = child.tail.strip("\n").strip()
                c_print(f"#[{w_col.wod_subtitle} bold]{child.text}", end=" ")
                c_print(f"#[{w_col.wod_subtitle} bold]{tail}", end="\n")

        elif tag == "p":
            print_wod_p(elm)

        elif tag == "div" and elm.attrib["class"] == "wotd-examples":
            child = elm.getchildren()[0].getchildren()[0]
            print_wod_p(child)


def print_wod_dyk(node):
    for elm in node.iterchildren():
        tag = elm.tag

        if tag == "h2":
            c_print(f"\n#[{w_col.wod_subtitle} bold]{elm.text}")

        elif tag == "p":
            print_wod_p(elm)


def parse_and_print_wod(res_url, res_text):
    logger.debug(f"{OP.PARSING.name} {res_url}")

    parser = etree.HTMLParser(remove_comments=True)
    tree = etree.HTML(res_text, parser)
    s = """
    //*[@class="article-header-container wod-article-header"] |
    //*[@class="wod-definition-container"] |
    //*[@class="did-you-know-wrapper"]
    """

    nodes = tree.xpath(s)

    logger.debug(f"{OP.PRINTING.name} the parsed result of {res_url}")

    print()
    for node in nodes:
        attr = node.attrib["class"]

        if "header" in attr:
            print_wod_header(node)

        elif "definition" in attr:
            print_wod_def(node)

        elif "did-you-know" in attr:
            print_wod_dyk(node)
    print()


async def parse_and_print_wod_calendar(session, res_url, res_text):
    logger.debug(f"{OP.PARSING.name} {res_url}")

    parser = etree.HTMLParser(remove_comments=True)
    tree = etree.HTML(res_text, parser)
    nodes = tree.xpath("//li/h2/a")
    logger.debug(f"{OP.PRINTING.name} the parsed result of {res_url}")

    data = {}
    for node in nodes:
        data[node.text] = node.attrib["href"]

    select_word = get_wod_selection_by_fzf(data) if has_tool("fzf") else get_wod_selection(data)
    if select_word in data.keys():
        url = WEBSTER_BASE_URL + data[select_word]
        await get_webster_wod_past(session, url)
    elif len(select_word) > 1 and not select_word.isnumeric():
        await search_webster(session, select_word)
    else:
        sys.exit()


async def get_webster_wod(session):
    resp = await fetch(session, WEBSTER_WORD_OF_THE_DAY_URL)
    result = await resp.text()
    parse_and_print_wod(resp.url, result)


async def get_webster_wod_past(session, req_url):
    resp = await fetch(session, req_url)
    result = await resp.text()
    parse_and_print_wod(resp.url, result)


async def get_webster_wod_list(session):
    resp = await fetch(session, WEBSTER_WORD_OF_THE_DAY_URL_CALENDAR)
    result = await resp.text()
    await parse_and_print_wod_calendar(session, resp.url, result)



async def main():
    try:
        async with aiohttp.ClientSession() as session:
            args = parse_args(session)
            args_dict = vars(args) # transfrom namespace object into a dict

            if args_dict.get("debug"):
                logger.setLevel(logging.DEBUG)
                logger.debug(args)

            if args_dict.get("subparser_name") is not None and args_dict.get("subparser_name") == "l":
                await list_words(args)
            elif args_dict.get("subparser_name") is not None and args_dict.get("subparser_name") == "wod":
                await wod(args)
            else:
                await search_word(args)

    except asyncio.exceptions.CancelledError:
        print("Task cancelled.")
        try:
            sys.exit(2)
        except SystemExit:
            pass

    #except KeyboardInterrupt:
    #    print("\nStopped by user.")

    except SystemExit:
        pass

    con.close()


def run_on_term():
    asyncio.run(main())


if __name__ == "__main__":
    import os
    import sys
    import asyncio
    import aiohttp # type: ignore
    import logging

    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

    from cambridge.args import parse_args, search_word, list_words, wod
    from cambridge.cache import con
    from cambridge.log import logger

    asyncio.run(main())

else:
    import sys
    import asyncio
    import aiohttp # type: ignore
    import logging

    from .args import parse_args, search_word, list_words, wod
    from .cache import con
    from .log import logger



import logging

# Logging only to stdout, not to local
# data = Path.home() / ".local" / "share" / "cambridge"
# data.mkdir(parents=True, exist_ok=True)

logging.basicConfig(
    # filename=str(data / "dict.log"),
    # filemode="a",
    # format="%(asctime)s - %(pathname)s[line:%(lineno)d] - %(levelname)s: %(message)s",
    format="%(asctime)s %(levelname)s %(name)s.%(filename)s[%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=logging.INFO,
)

logger = logging.getLogger(__package__)  # if imported, all modules will have the same logger called by name of __package__'s value'
